
// General Imports
import com.sun.opengl.util.GLUT;
import java.awt.*;
import java.io.*;

// JOGL Imports
import javax.media.opengl.glu.GLUquadric;
import javax.media.opengl.glu.GLU;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;

/**
 * Renderer Class
 *
 * @version $Id: Renderer.java 813 2008-06-03 16:31:35Z mzaczek $
 * @since CTS 1.0
 */
public class Renderer implements GLEventListener {

  /** GLU object. */
  private GLU glu = new GLU();
  
  /**
   * GLUquadric instance of a star sphere used for displaying a sphere textured
   * with stars to represent the star field.
   */
  private GLUquadric starSphere;
  
  /** Ambient light array. */
  private float lightAmbient[] = {0.08f, 0.08f, 0.08f, 0.0f};
  
  /** Diffuse light array. */
  private float lightDiffuse[] = {1.0f, 1.0f, 1.0f, 0.0f};
  
  /** Specular light array. */
  private float lightSpecular[] = {0.15f, 0.15f, 0.15f, 0.0f};
  
  /** Ambient/Diffuse array. */
  private float lightAmbientDiffuse[] = {0.65f, 0.65f, 0.65f, 0.0f};
  
  /** Position of the light. */
  private float lightPosition[] = {100.0f, 0.0f, 0.0f, 1.0f};
  
  /** Near plane clipping distance. */
  public double nearPlane = 0.001;
  
  /** Far plane clipping distance. */
  public double farPlane = 2000000.0;
  
  /** Zoom factor based on distance of eye to object. */
  public double zoomFactor = 0.01;
  
  /** Width. */
  private int _width;
  
  /** Height. */
  private int _height;
  
  /** Reset flag that when true will ask the display method to perform a reshape. */
  public boolean reset = false;
  
  public Vector3d vEye = new Vector3d(1.0, 1.0, 1.0);
  
  public Vector3d vLook = new Vector3d(-0.05, -0.05, -0.05);
  
  public Vector3d vUp = new Vector3d(0.0, 1.0, 0.0);
  
  public Vector3d vRight = new Vector3d(1.0, 0.0, 0.0);
  
  public Point lastMousePoint = new Point();
  
  public Point currentMousePoint = new Point();
  
  public Camera camera = new Camera();

  /**
   * Renderer constructor that initializes the rotation strings and sets the 
   * width/height.
   * 
   * @param width The width.
   * @param height The height.
   */
  Renderer(int width, int height) {
    this._width = width;
    this._height = height;
    //camera.setEye(new Vector3d(5, 0, 0));
    //camera.lookAt(new Vector3d(0, 0, 0), new Vector3d(0, 0, 1));
    
    camera.setEye(new Vector3d(0, 0, -2));
    camera.lookAt(new Vector3d(0, 0, 0), new Vector3d(1, 0, 0));    
  }

  /**
   * The initialization function that configures the view, defines the lights,
   * loads the star and sun textures, and initializes the system and its
   * children.
   *
   * @param drawable A GLAutoDrawable instance.
   */
  public void init(GLAutoDrawable drawable) {
    final GL gl = drawable.getGL();

    gl.setSwapInterval(1);

    //
    // View Settings
    //
    gl.glViewport(0, 0, 320, 240);
    gl.glMatrixMode(GL.GL_PROJECTION);
    gl.glLoadIdentity();

    // Set the Field-of-View, Aspect Ratio, Near & Far clipping planes
    glu.gluPerspective(65.0, 320.0 / 240.0, nearPlane, farPlane);
    gl.glMatrixMode(GL.GL_MODELVIEW);
    gl.glLoadIdentity();

    //
    // Lighting
    //
    // Enable Smooth Shading
    gl.glShadeModel(GL.GL_SMOOTH);
    gl.glMaterialfv(gl.GL_FRONT, gl.GL_SPECULAR, lightSpecular, 0);
    gl.glMaterialfv(gl.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, lightAmbientDiffuse, 0);
    // Background - black
    gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    // Depth Buffer Setup
    gl.glClearDepth(1.0f);
    // Clear The Stencil Buffer To 0
    gl.glClearStencil(0);
    // Enables Depth Testing
    gl.glEnable(GL.GL_DEPTH_TEST);
    // Really Nice Perspective Calculations
    gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);
    // Set The Ambient Lighting For Light0
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, lightAmbient, 0);
    // Set The Diffuse Lighting For Light0
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, lightDiffuse, 0);
    // Set The Position For Light0
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition, 0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, lightSpecular, 0);
    gl.glLightf(GL.GL_LIGHT0, GL.GL_CONSTANT_ATTENUATION, 1.0f);
    gl.glLightf(GL.GL_LIGHT0, GL.GL_LINEAR_ATTENUATION, 0.0f);
    gl.glLightf(GL.GL_LIGHT0, GL.GL_QUADRATIC_ATTENUATION, 0.0f);
    //gl.glLightModeli(GL.GL_LIGHT_MODEL_LOCAL_VIEWER, GL.GL_FALSE);
    gl.glLightModelfv(gl.GL_LIGHT_MODEL_AMBIENT, lightAmbient, 0);
    // Enable Lighting
    gl.glEnable(GL.GL_LIGHTING);
    // Enable Light 0
    gl.glEnable(GL.GL_LIGHT0);
    gl.glEnable(GL.GL_CULL_FACE);
    gl.glEnable(GL.GL_NORMALIZE);
    gl.glLightModelf(GL.GL_LIGHT_MODEL_COLOR_CONTROL, GL.GL_SEPARATE_SPECULAR_COLOR);

    // Enable lighting on both side...especially needed for certain models
    // where the normal calculations would give the wrong face.
    // Note that this failed for some models being loaded...test again.
    //gl.glLightModeli(GL.GL_LIGHT_MODEL_TWO_SIDE,1);
  }

  /**
   * The primary display function that is called by JOGL to draw the defined
   * scene - it is called for every frame.  It first clears the background of
   * the previous scene.  Next the rotation parameters are corrected/updated.
   * The star background is then displayed, which is then followed by the
   * display of the planets and models attached to the system.  And finally,
   * the sun texture is drawn and pointed at the user.
   *
   * @param drawable A GLAutoDrawable instance.
   */
  public void display(GLAutoDrawable drawable) {
    GL gl = drawable.getGL();

    if (reset) {
      int width = _width;
      int height = _height;
      gl.glViewport(0, 0, width, height);
      gl.glMatrixMode(GL.GL_PROJECTION);
      gl.glLoadIdentity();
      // Set the Field-of-View, Aspect Ratio, Near & Far clipping planes
      glu.gluPerspective(65.0, (double) width / (double) height, nearPlane, farPlane);
      gl.glMatrixMode(GL.GL_MODELVIEW);
      gl.glLoadIdentity();
      reset = false;
    }

    // Clear Screen And Depth Buffer
    gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
    gl.glLoadIdentity();

    updateView(gl);

    // Update light position
    // 0 in the last component causes jagged terminator on planets, 1 results
    // in a smooth terminator but the location of the sunlight is no longer
    // correct...not sure why.  Old code used a 0 (zero) very successfully,
    // what changed?!
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition, 0);

    GLUT glut = new GLUT();
    
    gl.glEnable(GL.GL_LIGHTING);
    gl.glEnable(GL.GL_COLOR_MATERIAL);    
    
    // Draw a large sphere
    glut.glutWireSphere(1000.0, 32, 32);
    
    // Draw Teapot
    gl.glDisable(GL.GL_CULL_FACE);    
    gl.glColor3f(1f, 1f, 0f);
    glut.glutSolidTeapot(0.5);
    gl.glDisable(GL.GL_LIGHTING);
    gl.glColor3f(1f, 1f, 1f);
    glut.glutWireTeapot(0.505);
    gl.glScaled(1.0, 1.0, 1.0);
    displayAxes(gl);
    
    
      gl.glPushMatrix();
        gl.glColor3f(1f, 0f, 0f);
        gl.glTranslated(5.0, 0.0, 0.0);
        glut.glutWireSphere(0.1, 16, 16);
        gl.glColor3f(1f, 1f, 1f);
      gl.glPopMatrix();  
      gl.glPushMatrix();
        gl.glColor3f(0f, 1f, 0f);
        gl.glTranslated(0.0, 5.0, 0.0);
        glut.glutWireSphere(0.1, 16, 16);
        gl.glColor3f(1f, 1f, 1f);
      gl.glPopMatrix();  
      gl.glPushMatrix();
        gl.glColor3f(0f, 0f, 1f);
        gl.glTranslated(0.0, 0.0, 5.0);
        glut.glutWireSphere(0.1, 16, 16);
        gl.glColor3f(1f, 1f, 1f);
      gl.glPopMatrix();     
      gl.glPushMatrix();
        gl.glColor3f(1f, 1f, 1f);
        gl.glTranslated(5.0, 5.0, 5.0);
        glut.glutWireSphere(0.1, 16, 16);
        gl.glColor3f(1f, 1f, 1f);
      gl.glPopMatrix();  
    
    // Only display the rotation point when in 'orbit' mode.
    if (camera.getOrbit()) {
      gl.glPushMatrix();
        gl.glColor3f(1f, 0f, 1f);
        gl.glTranslated(camera.getOrbitPointCopy().x, camera.getOrbitPointCopy().y, camera.getOrbitPointCopy().z);
        glut.glutWireSphere(0.1, 16, 16);
        gl.glColor3f(1f, 1f, 1f);
      gl.glPopMatrix();
//
//      gl.glBegin(gl.GL_LINES);
//        gl.glVertex3d(0, 0, 0);
//        gl.glVertex3d(camera.getEye().x, camera.getEye().y, camera.getEye().z);
//      gl.glEnd();
    }
    
    // Flush The GL Rendering Pipeline
    gl.glFlush();
  }

  /**
   * This method affects the display if the window is reshaped.
   *
   * @param drawable A GLAutoDrawable instance.
   * @param x The x coordinate of the top-left corner of the window.
   * @param y The y coordinate of the top-left corner of the window.
   * @param width The width of the window.
   * @param height The height of the window.
   */
  public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
    GL gl = drawable.getGL();

    // Prevent division by 0 in aspect ratio
    if (height <= 0) {
      height = 1;
    }

    this._height = height;
    this._width = width;
    gl.glViewport(0, 0, width, height);
    gl.glMatrixMode(GL.GL_PROJECTION);
    gl.glLoadIdentity();
    // Set the Field-of-View, Aspect Ratio, Near & Far clipping planes
    glu.gluPerspective(65.0, (double) width / (double) height, nearPlane, farPlane);
    gl.glMatrixMode(GL.GL_MODELVIEW);
    gl.glLoadIdentity();
  }

  /**
   * This method is not completed in the current JOGL specification (1.1.0)
   * but is is required because of the implementation specification.
   *
   * @param drawable A GLAutoDrawable instance.
   * @param modeChanged Flag for mode change.
   * @param deviceChanged Flag for device change (ie if window is moved to another
   *     screen, etc).
   */
  public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {
    // Nothing
  }
    
  private void updateView(GL gl) {  
    
    
    gl.glMultMatrixd(camera.getView(), 0);            

/// updateViewMatrix(gl);
  //          glu.gluLookAt( vEye.x,    vEye.y,    vEye.z,
  //                        vLook.x,   vLook.y,   vLook.z,
  //                          vUp.x,     vUp.y,     vUp.z);          
  }

  private void displayAxes(GL gl) {
    gl.glDisable(GL.GL_TEXTURE_2D);
    gl.glDisable(GL.GL_LIGHTING);
    gl.glDisable(GL.GL_COLOR_MATERIAL);
    gl.glLineWidth(3.0f);

    // Display red colored X axis
    gl.glColor3d(1.0, 0.0, 0.0);
    gl.glBegin(gl.GL_LINES);
    gl.glVertex3d(0.0, 0.0, 0.0);
    gl.glVertex3d(1.0, 0.0, 0.0);
    gl.glEnd();

    // Display green colored Y axis
    gl.glColor3d(0.0, 1.0, 0.0);
    gl.glBegin(gl.GL_LINES);
    gl.glVertex3d(0.0, 0.0, 0.0);
    gl.glVertex3d(0.0, 1.0, 0.0);
    gl.glEnd();

    // Display blue colored Z axis
    gl.glColor3d(0.0, 0.0, 1.0);
    gl.glBegin(gl.GL_LINES);
    gl.glVertex3d(0.0, 0.0, 0.0);
    gl.glVertex3d(0.0, 0.0, 1.0);
    gl.glEnd();

    gl.glLineWidth(1.0f);
    gl.glColor3d(1.0, 1.0, 1.0);
  }

  /**
   * Builds a view matrix.
   * 
   * The final view matrix will take the form of:
   * <pre>
   *  |   rx    ry    rz   -(r.e)  |
   *  |   ux    uy    uz   -(u.e)  |
   *  |  -lx   -ly   -lz    (l.e)  |
   *  |    0     0     0       1   |
   * 
   * Where r = Right vector
   *       u = Up vector
   *       l = Look vector
   *       e = Eye position in world space
   *       . = Dot-Product operation
   * </pre>
   * 
   */
  private void updateViewMatrix(GL gl) {
    double view[] = {1.0, 0.0, 0.0, 0.0,
      0.0, 1.0, 0.0, 0.0,
      0.0, 0.0, 1.0, 0.0,
      0.0, 0.0, 0.0, 1.0
    };
    vLook.normalize();

    Vector3d.cross(vRight, vLook, vUp);
    vRight.normalize();

    Vector3d.cross(vUp, vRight, vLook);
    vUp.normalize();

    view[0] = vRight.x;
    view[1] = vUp.x;
    view[2] = -vLook.x;
    view[3] = 0.0;

    view[4] = vRight.y;
    view[5] = vUp.y;
    view[6] = -vLook.y;
    view[7] = 0.0;

    view[8] = vRight.z;
    view[9] = vUp.z;
    view[10] = -vLook.z;
    view[11] = 0.0;

    view[12] = -Vector3d.dot(vRight, vEye);
    view[13] = -Vector3d.dot(vUp, vEye);
    view[14] = Vector3d.dot(vLook, vEye);
    view[15] = 1.0;

//    System.out.println("  vEye: " + vEye);
//    System.out.println(" vLook: " + vLook);
//    System.out.println("   vUp: " + vUp);
//    System.out.println("vRight: " + vRight);

    gl.glMultMatrixd(view, 0);
  }

  public double[] matRotate(double mat[], double angle, Vector3d axis) {
    double s = Math.sin(angle * Math.PI / 180.0);
    double c = Math.cos(angle * Math.PI / 180.0);

    axis.normalize();

    mat[0] = c + (1 - c) * axis.x;
    mat[1] = (1 - c) * axis.x * axis.y + s * axis.z;
    mat[2] = (1 - c) * axis.x * axis.z - s * axis.y;
    mat[3] = 0.0;

    mat[4] = (1 - c) * axis.y * axis.x - s * axis.z;
    mat[5] = c + (1 - c) * Math.pow(axis.y, 2);
    mat[6] = (1 - c) * axis.y * axis.z + s * axis.x;
    mat[7] = 0.0;

    mat[8] = (1 - c) * axis.z * axis.x + s * axis.y;
    mat[9] = (1 - c) * axis.z * axis.z - s * axis.x;
    mat[10] = c + (1 - c) * Math.pow(axis.z, 2);
    mat[11] = 0.0;

    mat[12] = 0.0;
    mat[13] = 0.0;
    mat[14] = 0.0;
    mat[15] = 1.0;

    return mat;
  }

  public Vector3d transformVector(double mat[], Vector3d vec) {
    double x = vec.x;
    double y = vec.y;
    double z = vec.z;

    vec.x = x * mat[0] +
            y * mat[4] +
            z * mat[8];

    vec.y = x * mat[1] +
            y * mat[5] +
            z * mat[9];

    vec.z = x * mat[2] +
            y * mat[6] +
            z * mat[10];

    return vec;
  }
  
  /**
   * Source: http://local.wasp.uwa.edu.au/~pbourke/geometry/rotate/
   * s
   * @param angle
   */
  public void rotateLookVector( double angle ) {    
    double a = angle * Math.PI/180.0;
    Quat4f q1 = new Quat4f((float)vUp.x, (float)vUp.y, (float)vUp.z, 0.0f);
    Quat4f q2 = new Quat4f((float)(vLook.x * Math.sin(a/2.0)),
                           (float)(vLook.y * Math.sin(a/2.0)),
                           (float)(vLook.z * Math.sin(a/2.0)),
                           (float)(Math.cos(a/2)));
    Quat4f q3 = new Quat4f();
    q3 = Quat4f.mult(q2, Quat4f.mult(q1, Quat4f.conjugate(q2)));
    
//    System.out.println("--------------angle: " + a + " ---------------------");
//    System.out.println("   vUp: " + vUp);
//    System.out.println("    q3: " + q3);
    
    vUp.x = q3.x;
    vUp.y = q3.y;
    vUp.z = q3.z;
  }
}